from SimulatorFramework import SimUtils,TileType,Tile,TileAssembly
from SimulatorFramework import Frontier
from PyTASTileType import *
import os
import sys
import random
import importlib



#===============================================================================
# N  = 0
# NE = 1
# E  = 2
# SE = 3
# S  = 4
# SW = 5
# W  = 6
# NW = 7
# ALLDIRS = 8
# DIRSTRS = ("N", "NE", "E", "SE", "S", "SW", "W", "NW")
#===============================================================================
#NBRS = [(0,1),(1,1),(1,0),(1,-1),(0,-1),(-1,-1),(-1,0),(-1,1)]

TILESET_LOAD_SUCCESS = 0
TILESET_LOAD_FAIL = 1
TILESET_LOAD_FAIL_NEEDS_3D = 2


#----------------------------------------------------------------------------
        
class PyTASAssembly(TileAssembly.Assembly):

    def __init__(self, tileWidths=100, tileHeights=100):
        TileAssembly.Assembly.__init__(self, tileWidths, tileHeights)
        
        self.NBRS = [(0,1),(1,0),(0,-1),(-1,0)]
        self.DIRSTRS = ["N", "E", "S", "W"]
        self.OPPDIRS = [2, 3, 0, 1]

        self.growthIncrement = 500
        self.temperature = 0
        self.FrontierHashTable = dict()
        self.bHas3D = False
        self.useScript = False
        self.cacheSearches = 0
        self.cacheHits = 0
        
        self.useMacrotileMode = False
        self.macrotileSize = None
        self.macrotileImageDir = None
        self.macrotileImages = dict()

    def ClearSpace(self):
        TileAssembly.Assembly.ClearSpace(self)
        self.FrontierHashTable = dict()
        self.cacheSearches = 0
        self.cacheHits = 0
    
    def GetFittingTilesForLoc(self, coords, growIfNeeded=True):
        #print("In GetFittingTilesForLoc for " + str(coords))
        retSet = None
        count = 0
        nbrGlues = []
        nbrGluesList = NeighborGlues()
        incidentStrengthTot = 0
        for n in self.NBRS:
            if self.bHas3D == True:
                nbrCoords = (coords[0]+n[0], coords[1]+n[1], coords[2]+n[2])
            else:
                nbrCoords = (coords[0]+n[0], coords[1]+n[1])
            nbrTile = self.GetLoc(nbrCoords, growIfNeeded)
            if not nbrTile is None:
                oppDir = self.OPPDIRS[count]
                nbrGlues.append(nbrTile.type.sides[oppDir].Hash())
                incidentStrengthTot = incidentStrengthTot + nbrTile.type.sides[oppDir].strength
                if self.useScript:
                    nbrGluesList.sides[count] = nbrTile.type.sides[oppDir]
                #if retSet is None:
                #    retSet = nbrTile.type.matches[oppDir]
                #else:
                #    retSet = retSet.intersection(nbrTile.type.matches[oppDir])
            else:
                nbrGlues.append(None)
                nbrGluesList.sides[count] = None
            count = count + 1
            
        # If there aren't strong enough incident glues to allow a tile to bind, get outta here
        if incidentStrengthTot < self.temperature:
            return None

        signature = str(nbrGlues)
        #self.cacheSearches = self.cacheSearches + 1
        if signature in self.FrontierHashTable.keys():
            #print("hit with signature " + signature)
            #self.cacheHits = self.cacheHits + 1
            return self.FrontierHashTable[signature]
        #else:
            #print("missed with signature " + signature)

        if self.useScript:
            userTileTypes = self.scriptModule.GetTileTypes(nbrGluesList)
            ## todo Check that supplied tiles can actually fit
            if (not userTileTypes is None) and (len(userTileTypes) > 0):
                if retSet is None:
                    retSet = set()
                for userTileType in userTileTypes:
                    userTileType.ParseColor()
                    if self.tileTypes.get(userTileType.name) == None:
                        #print("Adding name:" + userTileType.name)
                        self.tileTypes[userTileType.name] = userTileType
                    else:
                        #print("Name:" + userTileType.name + " is duplicate!")
                        raise ValueError("Name" + userTileType.name + " is duplicate!")
                        
                    retSet.add(userTileType.name)
                self.FrontierHashTable[signature] = retSet
        else:
            bDoNew = True
            
            if bDoNew == True:
                # Get the sets of tiles with matching glues for each direction
                ###print("nbr glues: " + str(nbrGlues))
                ngSets = []
                nCount = 0
                for ng in nbrGlues:
                    if (not ng is None) and (ng in self.tileSideHashtables[nCount].keys()):
                        ngSets.append((self.tileSideHashtables[nCount][ng], int(ng[-1])))
                        ###print("appending:" + str(self.tileSideHashtables[nCount][ng]) + ": and :" + str(int(ng[-1])) + ":")
                    #else:
                    #    ngSets.append(None)
                    nCount = nCount + 1
                #print("In GetFittingTilesForLoc for " + str(coords) + " got ngSets: " + str(ngSets))
                subsets = self.GetAllSubsets(ngSets)
                allFitters = set()
                for subset in subsets[1:]:
                    fitters = None
                    strength = 0
                    for side in subset:
                        strength = strength + side[1]
                    if strength >= self.temperature:
                        for side in subset:
                            if fitters is None:
                                fitters = set()
                                fitters = side[0]
                            else:
                                fitters = fitters & side[0]
                        if not fitters is None:
                            allFitters = allFitters | fitters
                #print("All fitting tile types in " + str(coords) + ": " + str(allFitters))
                if len(allFitters) > 0:
                    #print("Adding signature " + signature)
                    self.FrontierHashTable[signature] = allFitters
                    return allFitters
                else:
                    self.FrontierHashTable[signature] = None

            else:
                #Old, brute force method for finding fitting tile types
                for t in self.tileTypes:
                    totStr = 0
                    for n in range(len(self.NBRS)):
                        if not nbrGlues[n] is None:
                            if nbrGlues[n] == self.tileTypes[t].sides[n].Hash():
                                totStr = totStr + self.tileTypes[t].sides[n].strength
                    if totStr >= self.temperature:
                        if retSet is None:
                            retSet = set()
                        retSet.add(t)
                self.FrontierHashTable[signature] = retSet

        return retSet
        
    def LoadTileSystem(self, fileName, backgroundColor=None, bLoadTileset=True):
        if len(fileName) < 1:
            return False
        elif os.path.isabs(fileName) == False:
            fileName = "./" + fileName
            fileName = os.path.abspath(fileName)

        if bLoadTileset == True:
            self.tileTypes = dict()
        self.seedPatch = []
        self.frontier = set()
        self.temperature = 2    # Default temperature = 2
        self.bHas3D = False

        if self.useScript:
            directory = os.path.dirname(fileName)
            moduleName,fileExtension = os.path.splitext(os.path.split(fileName)[-1])
            print("directory:",directory,"moduleName:",moduleName,"fileExtension:",fileExtension)
            ## Import user script
            sys.path.insert(0, directory) # Add script's directory to path to find it by name
            self.scriptModule = importlib.import_module(moduleName) # Script can be referenced at self.scriptModule
            
            ## todo Check that required methods are defined
            
            ## todo Import any global settings defined by script (e.g. temperature)
            try:
                self.temperature = self.scriptModule.TEMPERATURE
                #print("Set temperature", self.temperature)
            except AttributeError:
                print("No temperature defined in script, using default:", self.temperature)
            
            ## Get seed tiles and positions from user script
            seedTiles = self.scriptModule.GetSeed()
            for seed in seedTiles:
                seedTType = seed[0] #Get the PyTASTileType object from the script output
                seedLocation = seed[1]  #Get the position
                seedTType.ParseColor()
                if self.tileTypes.get(seedTType.name) == None:
                    self.tileTypes[seedTType.name] = seedTType  #Add to tiles list
                    self.seedPatch.append((seedTType.name, seedLocation))   #Add to seed list
                else:
                    raise ValueError("Name" + seedTType.name + " is duplicate!")
                ## todo? change name like in LoadTileSet() ?
            SimUtils.globalApp.OutputMessage("Done loading script")
        else:
            #try:
            os.stat(fileName)
            rFile = open(fileName, "r")
            SimUtils.globalApp.OutputMessage("Using tile system definition file '" + fileName + "'")
            lines = rFile.readlines()
            rFile.close()

            self.systemDefFile = fileName
            
            if bLoadTileset == True:
                tdsFileName = lines[0].strip()
                if os.path.isabs(tdsFileName) == False:
                    tdsFileName = os.path.dirname(fileName) + "/" + tdsFileName
                    tdsFileName = os.path.abspath(tdsFileName)
                tilesetLoadRet = self.LoadTileSet(tdsFileName, backgroundColor=backgroundColor)
                if not tilesetLoadRet == TILESET_LOAD_SUCCESS:
                    return tilesetLoadRet
    
            for line in lines[1:]:
                if line.startswith("Temperature="):
                    temperatureStr = line[len("Temperature="):]
                    self.temperature = int(temperatureStr)
                elif line.startswith("MacrotileSize="):
                    tempStr = line[len("MacrotileSize="):]
                    self.macrotileSize = int(tempStr)
                    self.useMacrotileMode = True
                elif line.startswith("MacrotileImageDir="):
                    self.macrotileImageDir = line[len("MacrotileImageDir="):].strip()
                else:
                    seedLocArr = line.split(" ")
                    seedTileName = seedLocArr[0]
                    seedTileXLoc = int(seedLocArr[1])
                    seedTileYLoc = int(seedLocArr[2])
                    seedTileZLoc = 0
                    if len(seedLocArr) > 3:
                        seedTileZLoc = int(seedLocArr[3])
                        if not seedTileZLoc == 0:
                            self.bHas3D = True
                        self.seedPatch.append((seedTileName, (seedTileXLoc, seedTileYLoc, seedTileZLoc)))
                    else:
                        self.seedPatch.append((seedTileName, (seedTileXLoc, seedTileYLoc)))

            if self.useMacrotileMode == True:
                SimUtils.globalApp.OutputMessage("Using macrotile mode, with macrotile size " + str(self.macrotileSize))
                
            # If seed had 3D coords, make sure all tile types are 3D
            if self.bHas3D == True:
                for ttName in self.tileTypes:
                    self.tileTypes[ttName].SetDimensions(3)
                    
            self.tileSystemDef = dict()
            if bLoadTileset == True:
                self.tileSystemDef['tdsFile'] = tdsFileName
            self.tileSystemDef['temperature'] = self.temperature

        # Load tile set, then verify that tile types referenced in the seed actually exist
        midPt = self.PlaceSeed()
        return midPt
        
    def CreateTileTypeImages(self, backgroundColor=None):
        for ttypeName in self.tileTypes:
            self.tileTypes[ttypeName].MakeImage(backgroundColor=backgroundColor)
        
    def LoadTileSet(self, tdsFileName, backgroundColor=None):
        SimUtils.globalApp.OutputMessage("Reading tile set from '" + tdsFileName + "'")
        os.stat(tdsFileName)
        rFile = open(tdsFileName, "r")
        lines = rFile.readlines()
        rFile.close()

        i = 0
        ttype = PyTASTileType()
        while i < len(lines):
            lineArr = lines[i].strip().split(" ")
            if lineArr[0] == "TILENAME":
                ttype.name = lineArr[1]
            elif lineArr[0] == "LABEL" and len(lineArr) > 1:
                ttype.label = lineArr[1]
            elif lineArr[0] == "TILECOLOR" and len(lineArr) > 1:
                ttype.color = lines[i][len("TILECOLOR "):].strip()
            elif lineArr[0] == "TEXTCOLOR" and len(lineArr) > 1:
                ttype.textColor = lines[i][len("TEXTCOLOR "):].strip()
            elif lineArr[0] == "NORTHLABEL" and len(lineArr) > 1:
                ttype.sides[N].label = lineArr[1]
                if ttype.sides[N].label == "-99":
                    ttype.sides[N].strength = 0
                    ttype.sides[N].label = ""
            elif lineArr[0] == "EASTLABEL" and len(lineArr) > 1:
                ttype.sides[E].label = lineArr[1]
                if ttype.sides[E].label == "-99":
                    ttype.sides[E].strength = 0
                    ttype.sides[E].label = ""
            elif lineArr[0] == "SOUTHLABEL" and len(lineArr) > 1:
                ttype.sides[S].label = lineArr[1]
                if ttype.sides[S].label == "-99":
                    ttype.sides[S].strength = 0
                    ttype.sides[S].label = ""
            elif lineArr[0] == "WESTLABEL" and len(lineArr) > 1:
                ttype.sides[W].label = lineArr[1]
                if ttype.sides[W].label == "-99":
                    ttype.sides[W].strength = 0
                    ttype.sides[W].label = ""
            elif lineArr[0] == "UPLABEL" and len(lineArr) > 1:
                self.bHas3D = True
                ttype.SetDimensions(3)
                ttype.sides[U].label = lineArr[1]
                if ttype.sides[U].label == "-99":
                    ttype.sides[U].strength = 0
                    ttype.sides[U].label = ""
            elif lineArr[0] == "DOWNLABEL" and len(lineArr) > 1:
                self.bHas3D = True
                ttype.SetDimensions(3)
                ttype.sides[D].label = lineArr[1]
                if ttype.sides[D].label == "-99":
                    ttype.sides[D].strength = 0
                    ttype.sides[D].label = ""
            elif lineArr[0] == "NORTHBIND" and len(lineArr) > 1:
                ttype.sides[N].strength = int(lineArr[1])
            elif lineArr[0] == "EASTBIND" and len(lineArr) > 1:
                ttype.sides[E].strength = int(lineArr[1])
            elif lineArr[0] == "SOUTHBIND" and len(lineArr) > 1:
                ttype.sides[S].strength = int(lineArr[1])
            elif lineArr[0] == "WESTBIND" and len(lineArr) > 1:
                ttype.sides[W].strength = int(lineArr[1])
            elif lineArr[0] == "UPBIND" and len(lineArr) > 1:
                self.bHas3D = True
                ttype.SetDimensions(3)
                ttype.sides[U].strength = int(lineArr[1])
            elif lineArr[0] == "DOWNBIND" and len(lineArr) > 1:
                self.bHas3D = True
                ttype.SetDimensions(3)
                ttype.sides[D].strength = int(lineArr[1])
            elif lineArr[0] == "CREATE":
                ttype.ParseColor()
                if self.tileTypes.get(ttype.name) == None:
                    self.tileTypes[ttype.name] = ttype
                elif self.tileTypes.get(ttype.name) != None:
                    SimUtils.globalApp.OutputMessage("Warning, Duplicate tile name: " + ttype.name + "\tThis may cause unexpected behavior")
                    duplicateNum = 1
                    while self.tileTypes.get(ttype.name + "(" + str(duplicateNum) + ")") != None:
                        duplicateNum += 1
                    ttype.name = ttype.name + "(" + str(duplicateNum) + ")"
                    self.tileTypes[ttype.name] = ttype
                ttype = PyTASTileType()
            i = i + 1

        if (self.bHas3D == True) and (len(self.NBRS) == 4):
            return TILESET_LOAD_FAIL_NEEDS_3D
        
        if self.bHas3D == True:
            for ttName in self.tileTypes:
                self.tileTypes[ttName].SetDimensions(3)

        print("Building glue label hash tables to speed assembly simulation...")
        self.FrontierHashTable = dict()
        self.MakeTileSideHashtables()
        print(" ...done building glue label hash tables.")
        
        self.tilesetDefFile = tdsFileName

        # MJP: This greatly slows the loading of large tile sets, but increases
        #      the initial speed of assembly until all tile type images have been
        #      created and cached
        #self.CreateTileTypeImages(backgroundColor=backgroundColor)

        SimUtils.globalApp.OutputMessage("Loaded tile set with size: " + str(len(self.tileTypes)))

        return TILESET_LOAD_SUCCESS

    def SaveAssembly(self, fileName):

        #TODO: figure out what to do if self.useScript == True

        wFile = open(fileName, "w")
        
        # Make tds file name relative to tdp file name
        absoluteTdpDir = os.path.abspath(os.path.dirname(fileName))
        absoluteTdsDir = os.path.abspath(os.path.dirname(self.tilesetDefFile))
        commonPrefix = os.path.commonprefix([absoluteTdpDir, absoluteTdsDir])
        relPath = os.path.relpath(absoluteTdsDir, absoluteTdpDir)
        
        newTdsPath = os.path.join(relPath, os.path.basename(self.tilesetDefFile))
        print("newTdsPath: " + newTdsPath)
        
        # First write system variables
        wFile.write(newTdsPath + "\n")
        wFile.write("Temperature=" + str(self.temperature) + "\n")
        if (not self.macrotileSize is None) and (self.macrotileSize > 0):
            wFile.write("MacrotileSize=" + int(self.macrotileSize) + "\n")
        if (not self.macrotileImageDir is None) and len(self.macrotileImageDir) > 0:
            wFile.write("MacrotileImageDir=" + self.macrotileImageDir + "\n")
        
        # Now write out all tile locations
        for i in range(self.minX, self.maxX+1):
            for j in range(self.minY, self.maxY+1):
                loc = self.GetLoc((i,j))
                if loc is not None:
                    if self.bHas3D == True:
                        wFile.write(loc.type.GetName() + " " + str(loc.coords[0]) + " " + str(loc.coords[1]) + " " + str(loc.coords[2]) + "\n")
                    else:
                        wFile.write(loc.type.GetName() + " " + str(loc.coords[0]) + " " + str(loc.coords[1]) + "\n")

        wFile.close()
        SimUtils.globalApp.OutputMessage("Saved assembly to file '" + fileName + "'")
        self.systemDefFile = fileName

        return True
    
    def RebuildFrontier(self):
        self.FrontierHashTable = dict()

        self.frontier = set()
        savedMinX = self.minX
        savedMaxX = self.maxX
        savedMinY = self.minY
        savedMaxY = self.maxY
        for i in range(savedMinX, savedMaxX+1):
            for j in range(savedMinY, savedMaxY+1):
                if self.GetLoc((i,j), False, False) is None:
                    if self.AddToFrontierIfNecessary((i,j), growIfNeeded=False) is None:
                        self.ClearLoc((i,j))
                    
